@@ -1907,7 +1907,50 @@ wait_for_new_connections(int *fds, struct timeval *timeout, SockAddr *saddr)
 	 */
 	if (SERIALIZE_ACCEPT)
 	{
-		pool_semaphore_lock(ACCEPT_FD_SEM);
+		if (pool_config->connection_life_time == 0)
+		{
+			pool_semaphore_lock(ACCEPT_FD_SEM);
+		}
+		else
+		{
+			int sts;
+
+			for (;;)
+			{
+				sts = pool_semaphore_lock_allow_interrupt(ACCEPT_FD_SEM);
+
+				/* Interrupted by alarm */
+				if (sts == -2)
+				{
+					/*
+					 * Check if there are expired connection_life_time.
+					 */
+					if (backend_timer_expired)
+					{
+						/*
+						 * We add 10 seconds to connection_life_time so that there's
+						 * enough margin.
+						 */
+						int	seconds = pool_config->connection_life_time + 10;
+
+						while (seconds-- > 0)
+						{
+							/* check backend timer is expired */
+							if (backend_timer_expired)
+							{
+								pool_backend_timer();
+								backend_timer_expired = 0;
+								break;
+							}
+							sleep(1);
+						}
+					}
+				}
+				else	/* success or other error */
+					break;
+			}
+		}
+
 		set_ps_display("wait for connection request", false);
 		ereport(DEBUG1,
 				(errmsg("LOCKING select()")));
@@ -2138,7 +2181,7 @@ validate_backend_connectivity(int front_end_fd)
 		if (front_end_fd > 0)
 		{
 			POOL_CONNECTION *cp;
-			StartupPacket *volatile sp;
+			StartupPacket *volatile sp = NULL;
 
 			/*
 			 * we do not want to report socket error, as above errors will be
@@ -2259,6 +2302,7 @@ retry_startup:
 	{
 		cancel_request((CancelPacket *) sp->startup_packet);
 		pool_free_startup_packet(sp);
+		connection_count_down();
 		return NULL;
 	}
 
@@ -2274,6 +2318,20 @@ retry_startup:
 		goto retry_startup;
 	}
 
+	/* GSSAPI? */
+	if (sp->major == 1234 && sp->minor == 5680)
+	{
+		ereport(DEBUG1,
+				(errmsg("selecting backend connection"),
+				 errdetail("GSSAPI request from client")));
+
+		/* sorry, Pgpool-II does not support GSSAPI yet */
+		pool_write_and_flush(frontend, "N", 1);
+
+		pool_free_startup_packet(sp);
+		goto retry_startup;
+	}
+
 	frontend->protoVersion = sp->major;
 	frontend->database = pstrdup(sp->database);
 	frontend->username = pstrdup(sp->user);
